﻿//---------------------------------------------------------------------------
//
//   ExportASC
//
//          Copyright(C) 1999-2023, tetraface Inc.
//
//     Sample of Export plug-in for exporting an ASC (3D Studio ASCII) file.
//     Created DLL must be installed in "Plugins\Export" directory.
//     3DS is defined for an example of multiple formats, but this is just 
//    a sample. It is not implemented actually.
//
//    　ASC(3D Studio ASCII)形式を出力するエクスポートプラグインのサンプル。
//    　作成したDLLは"Plugins\Export"フォルダに入れる必要がある。
//    　ASC形式自体がサポートする機能が低いのと、加えて仕様の詳細がわからない
//    ため、このサンプルではあまり多くの情報を出力していない。
//    　複数の形式を同時サポートする例として3DSを使っているが、あくまで例なの
//    で実装はされていない。
//
//---------------------------------------------------------------------------

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <vector>
#include "MQPlugin.h"

BOOL SaveASC(const wchar_t *filename, MQDocument doc);


//---------------------------------------------------------------------------
//  DllMain
//---------------------------------------------------------------------------
BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
	//プラグインとしては特に必要な処理はないので、何もせずにTRUEを返す
    return TRUE;
}

//---------------------------------------------------------------------------
//  MQGetPlugInID
//    プラグインIDを返す。
//    この関数は起動時に呼び出される。
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT void MQGetPlugInID(DWORD *Product, DWORD *ID)
{
	// プロダクト名(制作者名)とIDを、全部で64bitの値として返す
	// 値は他と重複しないようなランダムなもので良い
	*Product = 0x56A31D20;
	*ID      = 0x45601E0C;
}

//---------------------------------------------------------------------------
//  MQGetPlugInName
//    プラグイン名を返す。
//    この関数は[プラグインについて]表示時に呼び出される。
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT const char *MQGetPlugInName(void)
{
	// プラグイン名
	return "3D Studio Exporter   Copyright(C) 1999-2023, tetraface Inc.";
}

//---------------------------------------------------------------------------
//  MQGetPlugInType
//    プラグインのタイプを返す。
//    この関数は起動時に呼び出される。
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT int MQGetPlugInType(void)
{
	// 出力用プラグインである
	return MQPLUGIN_TYPE_EXPORT;
}

//---------------------------------------------------------------------------
//  MQEnumFileType
//    入力または出力可能なファイルタイプを返す。
//    ファイルタイプは別名保存時のダイアログに表示される。
//    この関数は起動時に呼び出される。
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT const char *MQEnumFileType(int index)
{
	switch(index){
	case 0: return "3D Studio ASCII (*.asc)";
	// 複数のタイプをサポートする場合は以下のようにすればよいが、このサンプル
	// では実装されてないため、コメントアウト
	//case 1: return "3D Studio (*.3ds)"; 
	}
	return NULL;
}

//---------------------------------------------------------------------------
//  MQEnumFileExt
//    入力または出力可能な拡張子を返す。
//    この関数は起動時に呼び出される。
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT const char *MQEnumFileExt(int index)
{
	switch(index){
	case 0: return "asc";
	//case 1: return "3ds"; 
	}
	return NULL;
}

//---------------------------------------------------------------------------
//  MQExportFileW
//    [別名で保存]で保存するときに呼び出される。
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT BOOL MQExportFileW(int index, const wchar_t *filename, MQDocument doc)
{
	switch(index){
	case 0: return SaveASC(filename, doc);
	//case 1: return Save3DS(filename, doc);
	}
	return FALSE;
}


// ASC形式で出力
BOOL SaveASC(const wchar_t *filename, MQDocument doc)
{
	int i, face_index, mat_index;
	MQMaterial mat;
	MQColor color;
	MQFileDialogInfo dlginfo;

	// ファイルを作成・オープンする
	FILE *fh;
	errno_t ret_s = _wfopen_s(&fh, filename, L"w");
	if(ret_s != 0)
		return FALSE;

	// 座標軸変換用ダイアログの表示
	memset(&dlginfo, 0, sizeof(dlginfo));
	dlginfo.dwSize = sizeof(dlginfo);
	dlginfo.axis_x = MQFILE_TYPE_RIGHT;
	dlginfo.axis_y = MQFILE_TYPE_UP;
	dlginfo.axis_z = MQFILE_TYPE_BACK;
	dlginfo.softname = "3D Studio";
	MQ_ShowFileDialog("ASC Export", &dlginfo);

	// オブジェクトの数を得る
	int objcount = doc->GetObjectCount();

	// 複数のオブジェクトを１つに合成するために、作業用オブジェクトを作成する
	MQObject obj = MQ_CreateObject();

	// 作業用オブジェクトにすべてのオブジェクトを合成
	for(i=0; i<objcount; i++)
	{
		MQObject thisobj = doc->GetObject(i);
		if(thisobj == NULL) 
			continue;
		// フリーズなどにより内容が変更されてしまうので、オブジェクトを複製しておく
		thisobj = thisobj->Clone();
		// 曲面などの属性をフリーズ
		thisobj->Freeze(MQOBJECT_FREEZE_ALL);
		// 作業用オブジェクトに合成する
		obj->Merge(thisobj);
		// 複製したオブジェクトはもう不要なので破棄
		thisobj->DeleteThis();
	}

	// 面と頂点の数を取得
	int face_count = obj->GetFaceCount();
	int vert_count = obj->GetVertexCount();

	// ASC形式は三角形しか扱えないので、三角形面は１つ、四角形面は２つとして、
	// 出力後の三角面の数を数える
	int triangle_count = 0;
	for(face_index=0; face_index<face_count; face_index++)
	{
		int count = obj->GetFacePointCount(face_index);
		if(count >= 3){
			triangle_count += count-2;
		}
	}

	// ファイルの先頭にデータを書き込み
	fprintf(fh, "Ambient light color: Red=0.3 Green=0.3 Blue=0.3\n");
	fprintf(fh, "\n");
	fprintf(fh, "Named object: \"\"\n");
	fprintf(fh, "Tri-mesh, Vertices: %d     Faces: %d\n",
		vert_count, triangle_count);

	// ここから頂点リストの書き込み
	fprintf(fh, "Vertex list:\n");
	// 頂点のデータを配列としてptsに格納する
	MQPoint *pts = (MQPoint *)malloc(sizeof(MQPoint) * vert_count);
	obj->GetVertexArray(pts);
	// 座標軸をダイアログで指定したものに変換
	MQ_ExportAxis(&dlginfo, pts, vert_count);
	// ファイルへ出力
	for(i=0; i<vert_count; i++)
		fprintf(fh, "Vertex %d:  X:%.6f     Y:%.6f     Z:%.6f\n",
			i, pts[i].x, pts[i].y, pts[i].z);
	free(pts);

	// ここから面リストの書き込み
	fprintf(fh, "Face list:\n");
	triangle_count = 0;
	for(face_index=0; face_index<face_count; face_index++)
	{
		int count = obj->GetFacePointCount(face_index);
		if(count < 3) continue;

		// 頂点インデックスの配列を得る
		std::vector<int> index(count);
		obj->GetFacePointArray(face_index, &(*index.begin()));
		// 材質インデックスを得る(未着色面は-1が返される)
		mat_index = obj->GetFaceMaterial(face_index);
		mat = (mat_index >= 0) ? doc->GetMaterial(mat_index) : NULL;
		// 材質の色を得る(未着色面の場合は白色をセット)
		if(mat != NULL)
			color = mat->GetColor();
		else
			color = MQColor(1.0f, 1.0f, 1.0f);

		// 面を書き込む
		int tri_num = count-2;
		std::vector<int> tri_index(tri_num*3);
		if(count == 3){
			tri_index[0] = 0;
			tri_index[1] = 1;
			tri_index[2] = 2;
		}else{
			std::vector<MQPoint> pts(count);
			for(int i=0; i<count; i++){
				pts[i] = obj->GetVertex(index[i]);
			}
			doc->Triangulate(&(*pts.begin()), count, &(*tri_index.begin()), tri_num*3);
		}

		for(int i=0; i<tri_num; i++){
			fprintf(fh, "Face %d:    A:%d B:%d C:%d AB:1 BC:1 CA:1\n",
				triangle_count++, index[tri_index[0]], index[tri_index[1]], index[tri_index[2]]);
			fprintf(fh, "Material:\"r%dg%db%da%d\"\n", 
				(int)(color.r*255),(int)(color.b*255),(int)(color.b*255),255);
			fprintf(fh, "Smoothing:  1\n");
		}
	}

	// 作業用オブジェクトを破棄
	obj->DeleteThis();

	// ここからファイルの終端のデータを書き込み
	fprintf(fh, "\n");
	fprintf(fh, "									 Page 1\n");
	fprintf(fh, "\n");

	// ファイルをクローズ
	fclose(fh);

	return TRUE;
}

