
#include "FontUtil.h"
#include "StringUtil.h"
#include <codecvt>
#include <locale>

namespace FontUtil {
	
	static NSString* WideToNSString(const std::wstring& str)
	{
		return [[NSString alloc] initWithBytes:str.c_str() length:str.length()*sizeof(wchar_t) encoding:NSUTF32LittleEndianStringEncoding];
	}
	
	static NSFont* CreateNSFont(const std::wstring& fontname, int size)
	{
		NSString *fontstr = WideToNSString(fontname);
		NSFont *nsfont = nil;
		NSFont *sysfont = [NSFont systemFontOfSize:size];
		if([[sysfont fontName] compare:fontstr] == NSOrderedSame){
			// 同じ名前にしてもなぜか同じにならないのでシステムフォントを直接使う
			nsfont = sysfont;
		}else{
			nsfont = [NSFont fontWithName:fontstr size:size];
			if(nsfont == nil){
				nsfont = sysfont;
			}
		}
		return nsfont;
	}

	void EnumFonts(std::vector<std::pair<std::wstring, std::wstring>>& fonts)
	{
		NSArray *fontArray = [[NSFontManager sharedFontManager] availableFonts];
		for(NSString *str in fontArray){
			NSFont *font = [NSFont fontWithName:str size:20];
			if(font == nil)
				continue;
			NSString *disp = [font displayName];
			if(disp == nil)
				continue;
			fonts.push_back(std::make_pair(StringUtil::Utf8ToWide([str UTF8String]),
										   StringUtil::Utf8ToWide([disp UTF8String])));
		}
	}
	
	std::wstring GetApplicationFont()
	{
		NSFont *font = [NSFont userFontOfSize:[NSFont systemFontSize]];
		NSString *fontname = [font fontName];
		return StringUtil::Utf8ToWide([fontname UTF8String]);
	}
	
	float GetFontGlyph(const std::wstring& str, const std::wstring& fontname, std::vector<GlyphData>& data)
	{
		NSString *nsstr = WideToNSString(str);
		NSFont* font = CreateNSFont(fontname, 20);
		NSMutableDictionary *attr = [[NSMutableDictionary alloc] init];
		[attr setObject:font forKey:NSFontAttributeName];
		CGSize size = [nsstr sizeWithAttributes:@{NSFontAttributeName:font}];

		NSTextStorage *storage = [[NSTextStorage alloc] initWithString:nsstr attributes:attr];
		NSLayoutManager *manager = [[NSLayoutManager alloc] init];
		NSTextContainer *container = [[NSTextContainer alloc] init];
		[storage addLayoutManager:manager];
		[manager addTextContainer:container];
		
		NSUInteger length = [manager numberOfGlyphs];
		CGGlyph* glyphs = new CGGlyph[length+1];
		[manager getGlyphsInRange:NSMakeRange(0, length) glyphs:glyphs properties:nil characterIndexes:nil bidiLevels:nil];
		
		container = nil;
		manager = nil;
		storage = nil;
		
		NSBezierPath* path = [NSBezierPath bezierPath];
		[path moveToPoint: NSMakePoint(0, 0)];
		if(@available(macos 10.13, *)){
			[path appendBezierPathWithCGGlyphs:glyphs count:length inFont:font];
		}else{
			NSGlyph *nsglyphs = new NSGlyph[length+1];
			for(NSUInteger i=0; i<length; i++)
				nsglyphs[i] = glyphs[i];
			[path appendBezierPathWithGlyphs:nsglyphs count:length inFont:font];
			delete[] nsglyphs;
		}
		//NSPoint pt = [path currentPoint];
		//NSLog(@"glyph %f %f", pt.x, pt.y);
		
		delete[] glyphs;
		
		for(NSInteger i=0; i<[path elementCount]; ++i)
		{
			NSPoint controlPoints[3];
			NSBezierPathElement el = [path elementAtIndex:i associatedPoints:controlPoints];
			if(el == NSMoveToBezierPathElement)
			{
				data.push_back(GlyphData());
				data.back().type = kGlyphMove;
				data.back().point[0].x = controlPoints[0].x;
				data.back().point[0].y = controlPoints[0].y;
			}
			else if (el == NSLineToBezierPathElement)
			{
				data.push_back(GlyphData());
				data.back().type = kGlyphLine;
				data.back().point[0].x = controlPoints[0].x;
				data.back().point[0].y = controlPoints[0].y;
			}
			else if (el == NSCurveToBezierPathElement)
			{
				data.push_back(GlyphData());
				data.back().type = kGlyphCurve;
				data.back().point[0].x = controlPoints[0].x;
				data.back().point[0].y = controlPoints[0].y;
				data.back().point[1].x = controlPoints[1].x;
				data.back().point[1].y = controlPoints[1].y;
				data.back().point[2].x = controlPoints[2].x;
				data.back().point[2].y = controlPoints[2].y;
			}
			else if (el == NSClosePathBezierPathElement)
			{
				data.push_back(GlyphData());
				data.back().type = kGlyphClose;
			}
		}
		return size.width;
	}
	
}

